from __future__ import unicode_literals

from catsnap import Client, HASH_KEY
from boto.dynamodb.batch import BatchList
import json

def get_images(filenames):
    if not filenames:
        raise StopIteration
    dynamo = Client().get_dynamodb()
    table = Client().table('image')
    batch_list = BatchList(dynamo)
    batch_list.add_batch(table, filenames,
            attributes_to_get=['tags', HASH_KEY])
    response = dynamo.batch_get_item(batch_list)
    items = response['Responses'][table.name]['Items']
    for item in items:
        item['filename'] = item.pop(HASH_KEY)
        item['tags'] = json.loads(item['tags'])
    unprocessed_keys = []
    if response['UnprocessedKeys'] \
            and table.name in response['UnprocessedKeys']:
        for key in response['UnprocessedKeys'][table.name]['Keys']:
            unprocessed_keys.append(key['HashKeyElement'])

    for item in items:
        yield item
    if not unprocessed_keys:
        raise StopIteration
    for item in get_images(unprocessed_keys):
        yield item



from __future__ import unicode_literals

import json
from boto.dynamodb.exceptions import DynamoDBKeyNotFoundError

from catsnap.document import Document
from catsnap.ordered_set import OrderedSet

class Image(Document):
    _table_name = 'image'

    def __init__(self, filename, source_url=None):
        self.filename = filename
        self.source_url = source_url

    def add_tags(self, tags):
        existing_tags = []
        try:
            item = self._table().get_item(self.filename)
            existing_tags = json.loads(item['tags'])
        except DynamoDBKeyNotFoundError:
            item = self._table().new_item(hash_key=self.filename, attrs={})
        if self.source_url is not None:
            item['source_url'] = self.source_url
        item['tags'] = json.dumps(list(OrderedSet(existing_tags + tags)))
        item.put()

    def get_tags(self):
        try:
            item = self._table().get_item(self.filename)
        except DynamoDBKeyNotFoundError:
            return []
        return json.loads(item['tags'])

    def get_source_url(self):
        try:
            item = self._table().get_item(self.filename)
        except DynamoDBKeyNotFoundError:
            return None
        return item['source_url']



from __future__ import unicode_literals

from catsnap import Client, HASH_KEY
from boto.dynamodb.batch import BatchWriteList, BatchList
from boto.dynamodb.item import Item
import json

def get_tag_items(tag_names):
    if not tag_names:
        raise StopIteration

    dynamo = Client().get_dynamodb()
    table = Client().table('tag')
    batch_list = BatchList(dynamo)
    batch_list.add_batch(table, tag_names,
            attributes_to_get=['filenames', HASH_KEY])
    response = dynamo.batch_get_item(batch_list)
    items = response['Responses'][table.name]['Items']
    unprocessed_keys = []
    if response['UnprocessedKeys'] \
            and table.name in response['UnprocessedKeys']:
        for key in response['UnprocessedKeys'][table.name]['Keys']:
            unprocessed_keys.append(key['HashKeyElement'])

    for item in items:
        yield item
    for item in get_tag_items(unprocessed_keys):
        yield item

def get_tags(tag_names):
    for item in get_tag_items(tag_names):
        yield { 'tag': item[HASH_KEY],
                'filenames': json.loads(item['filenames'])}

def add_image_to_tags(filename, tag_names):
    items = list(get_tag_items(tag_names))
    for item in items:
        filenames = json.loads(item['filenames'])
        filenames.append(filename)
        item['filenames'] = json.dumps(filenames)

    table = Client().table('tag')
    for new_tag in set(tag_names) - set(x[HASH_KEY] for x in items):
        items.append(table.new_item(hash_key = new_tag,
                attrs={'filenames': json.dumps([filename])}))

    _submit_items(table, items)

def _submit_items(table, items):
    dynamo = Client().get_dynamodb()
    write_list = BatchWriteList(dynamo)
    write_list.add_batch(table, puts=items)
    response = write_list.submit()
    if 'UnprocessedItems' in response \
            and table.name in response['UnprocessedItems']:
        unprocessed_keys = set(x['PutRequest']['Item'][HASH_KEY]
                for x in response['UnprocessedItems'][table.name])
        unprocessed_items = filter(
                lambda x: x[HASH_KEY] in unprocessed_keys, items)
        _submit_items(table, unprocessed_items)



from __future__ import unicode_literals

from mock import patch, Mock, MagicMock
from nose.tools import eq_
from boto.dynamodb.exceptions import DynamoDBKeyNotFoundError
from tests import TestCase

from catsnap.document.tag import Tag
from catsnap import HASH_KEY

class TestAddingFile(TestCase):
    def test_sends_to_dynamo(self):
        item = MagicMock()
        table = Mock()
        table.get_item.side_effect = DynamoDBKeyNotFoundError('no such tag')
        table.new_item.return_value = item
        tag = Tag('cat')
        tag._stored_table = table

        tag.add_file('Sewing_cat.gif')
        table.new_item.assert_called_with(hash_key='cat', attrs ={})
        item.__setitem__.assert_called_with('filenames', '["Sewing_cat.gif"]')
        item.put.assert_called_with()

    def test_updates_existing_tag(self):
        table = Mock()
        item = MagicMock()
        item.__getitem__.return_value = '[ "Sewing_cat.gif" ]'
        table.get_item.return_value = item

        tag = Tag('cat')
        tag._stored_table = table
        tag.add_file('other_cat.gif')

        eq_(table.new_item.call_count, 0, "shouldn't've made a new entry")
        item.__setitem__.assert_called_with('filenames',
                '["Sewing_cat.gif", "other_cat.gif"]')
        item.put.assert_called_with()

    def test_appended_filenames_are_unique(self):
        table = Mock()
        item = MagicMock()
        item.__getitem__.return_value = '[ "Sewing_cat.gif" ]'
        table.get_item.return_value = item

        tag = Tag('cat')
        tag._stored_table = table
        tag.add_file('Sewing_cat.gif')
        item.__setitem__.assert_called_with('filenames', '["Sewing_cat.gif"]')

class TestGetFilenames(TestCase):
    def test_get_filenames(self):
        item = MagicMock()
        item.__getitem__.return_value = '["BADCAFE", "DEADBEEF"]'
        table = Mock()
        table.get_item.return_value = item
        tag = Tag('cat')
        tag._stored_table = table

        filenames = tag.get_filenames()
        eq_(filenames, ['BADCAFE', 'DEADBEEF'])
        table.get_item.assert_called_with('cat')

    def test_returns_empty_list_if_no_such_tag_exists(self):
        tag = Tag('cat')
        table = Mock()
        table.get_item.side_effect = DynamoDBKeyNotFoundError('no such tag')
        tag._stored_table = table

        eq_(tag.get_filenames(), [])



from __future__ import unicode_literals

#It's a big truck. You can just dump stuff on it.

import requests
import hashlib
import subprocess
import re

from catsnap import Client
from catsnap.config import MetaConfig

class ImageTruck():
    def __init__(self, contents, content_type, source_url):
        self.contents = contents
        self.content_type = content_type
        self.source_url = source_url

    @classmethod
    def new_from_url(cls, url):
        response = requests.get(url)
        response.raise_for_status()
        return cls(response.content, response.headers['content-type'],
                url)

    @classmethod
    def new_from_file(cls, filename):
        with open(filename, 'r') as image_file:
            contents = image_file.read()
        file_info = subprocess.check_output(['file', filename])
        match = re.search(r'(\w+) image data', file_info)
        if not match:
            raise TypeError("'%s' doesn't seem to be an image file" % filename)
        filetype = match.groups()[0].lower()

        return cls(contents, 'image/'+filetype, None)

    @classmethod
    def new_from_something(cls, path):
        url = requests.utils.urlparse(path)
        if url.scheme:
            return cls.new_from_url(path)
        else:
            return cls.new_from_file(path)

    @classmethod
    def url_for_filename(cls, filename):
        return cls._url(filename, MetaConfig().bucket,
                extension=MetaConfig().extension)

    def upload(self):
        key = Client().bucket().new_key(self.calculate_filename())
        key.set_metadata('Content-Type', self.content_type)
        key.set_contents_from_string(self.contents)
        key.make_public()

        filename = self.calculate_filename()

    def calculate_filename(self):
        return hashlib.sha1(self.contents).hexdigest()

    def url(self, **kwargs):
        return self._url(self.calculate_filename(), MetaConfig().bucket,
                extension=MetaConfig().extension)

    @classmethod
    def _url(cls, filename, bucket_name, extension=False):
        url = 'https://s3.amazonaws.com/%(bucket)s/%(filename)s' % {
                'bucket': bucket_name, 'filename': filename}
        if extension:
            url = url + '#.gif'
        return url



from __future__ import unicode_literals

from catsnap import Client

class Document():
    _stored_table = None
    _table_name = NotImplemented

    def _table(self):
        self._stored_table = self._stored_table \
                or Client().table(self._table_name)
        return self._stored_table



from __future__ import unicode_literals
import boto

from catsnap.config import MetaConfig
from catsnap.singleton import Singleton
from boto.exception import DynamoDBResponseError

#This really oughtta be, like, the tablename or something, but I screwed up, so
#now there're existing catsnap installs that use this schema. Sucks :(
#So yeah every table is keyed on an attribute called 'tag'
HASH_KEY = 'tag'

class Client(Singleton):
    _tables = {}
    _bucket = None

    _dynamo_connection = None
    _s3_connection = None

    def setup(self):
        created_tables = 0
        try:
            self.create_table('tag')
            created_tables += 1
        except DynamoDBResponseError, e:
            if e.error_code != 'ResourceInUseException':
                raise
        try:
            self.create_table('image')
            created_tables += 1
        except DynamoDBResponseError, e:
            if e.error_code != 'ResourceInUseException':
                raise

        return created_tables

    def bucket(self):
        if not self._bucket:
            bucket_name = MetaConfig().bucket
            s3 = self.get_s3()
            all_buckets = [x.name for x in s3.get_all_buckets()]
            if bucket_name not in all_buckets:
                self._bucket = s3.create_bucket(bucket_name)
            else:
                self._bucket = s3.get_bucket(bucket_name)
        return self._bucket

    def create_table(self, table_name):
        table_prefix = MetaConfig().bucket
        table_name = '%s-%s' % (table_prefix, table_name)

        dynamo = self.get_dynamodb()
        schema = dynamo.create_schema(hash_key_name='tag',
                hash_key_proto_value='S')
        return dynamo.create_table(name=table_name,
                schema=schema,
                read_units=3,
                write_units=5)

    def table(self, table_name):
        table_prefix = MetaConfig().bucket
        table_name = '%s-%s' % (table_prefix, table_name)

        if table_name not in self._tables:
            dynamo = self.get_dynamodb()
            self._tables[table_name] = dynamo.get_table(table_name)
        return self._tables[table_name]

    def get_dynamodb(self):
        if not self._dynamo_connection:
            self._dynamo_connection = boto.connect_dynamodb(
                    aws_access_key_id=MetaConfig().aws_access_key_id,
                    aws_secret_access_key=MetaConfig().aws_secret_access_key)
        return self._dynamo_connection

    def get_s3(self):
        if not self._s3_connection:
            self._s3_connection = boto.connect_s3(
                    aws_access_key_id=MetaConfig().aws_access_key_id,
                    aws_secret_access_key=MetaConfig().aws_secret_access_key)
        return self._s3_connection




from __future__ import unicode_literals

import json
from boto.dynamodb.exceptions import DynamoDBKeyNotFoundError

from catsnap.ordered_set import OrderedSet
from catsnap.document import Document

class Tag(Document):
    _table_name = 'tag'

    def __init__(self, name):
        self.name = name

    def add_file(self, filename):
        existing_filenames = []
        try:
            item = self._table().get_item(self.name)
            existing_filenames = json.loads(item['filenames'])
        except DynamoDBKeyNotFoundError:
            item = self._table().new_item(hash_key=self.name, attrs={})

        item['filenames'] = json.dumps(list(OrderedSet(existing_filenames +
                                                       [filename])))
        item.put()

    def get_filenames(self):
        try:
            item = self._table().get_item(self.name)
        except DynamoDBKeyNotFoundError:
            return []
        return json.loads(item['filenames'])



from __future__ import unicode_literals

from mock import patch, Mock, MagicMock, call
from nose.tools import eq_
from boto.dynamodb.exceptions import DynamoDBKeyNotFoundError
from tests import TestCase

from catsnap.document.image import Image

class TestAddTags(TestCase):
    def test_sends_to_dynamo(self):
        item = MagicMock()
        table = Mock()
        table.get_item.side_effect = DynamoDBKeyNotFoundError('no such image')
        table.new_item.return_value = item
        image = Image('deadbeef', 'mlkshk.com/kitty')
        image._stored_table = table

        image.add_tags(['cat', 'kitten'])
        table.new_item.assert_called_with(hash_key='deadbeef', attrs={})
        item.__setitem__.assert_has_calls([
                call('source_url', 'mlkshk.com/kitty'),
                call('tags', '["cat", "kitten"]')])
        item.put.assert_called_with()

    def test_appends_tags_to_existing_image(self):
        table = Mock()
        item = MagicMock()
        item.__getitem__.return_value = '["cat"]'
        table.get_item.return_value = item

        image = Image('BABB1E5')
        image._stored_table = table
        image.add_tags(['kitten'])

        eq_(table.new_item.call_count, 0, "shouldn't've made a new entry")
        item.__setitem__.assert_called_with('tags', '["cat", "kitten"]')
        item.put.assert_called_with()

    def test_appended_tags_are_unique(self):
        table = Mock()
        item = MagicMock()
        item.__getitem__.return_value = '["cat"]'
        table.get_item.return_value = item

        image = Image('BABB1E5')
        image._stored_table = table
        image.add_tags(['cat'])

        item.__setitem__.assert_called_with('tags', '["cat"]')

    def test_overwrites_source_url_if_present(self):
        table = Mock()
        item = MagicMock()
        def getitem(key):
            if key == 'tags':
                return '["cat"]'
            elif key == 'source_url':
                return 'http://mlkshk.com'
            else:
                raise KeyError(key)
        item.__getitem__.side_effect = getitem
        table.get_item.return_value = item
        image_with_url = Image('BABB1E5', 'http://imgur.com')
        image_with_url._stored_table = table

        image_with_url.add_tags(['kitten'])
        item.__setitem__.assert_has_calls([
            call('source_url', 'http://imgur.com'),
            call('tags', '["cat", "kitten"]')])

    def test_get_tags(self):
        table = Mock()
        item = MagicMock()
        item.__getitem__.return_value = '["cat", "kittycat", "dancedance"]'
        table.get_item.return_value = item
        image = Image('BADBADBADBADBAD')
        image._stored_table = table

        tags = image.get_tags()
        eq_(tags, ['cat', 'kittycat', 'dancedance'])

    def test_get_tags__returns_none_if_no_such_image(self):
        table = Mock()
        table.get_item.side_effect = DynamoDBKeyNotFoundError('no such image')
        image = Image("TH15I5N0TEVENHEX")
        image._stored_table = table

        tags = image.get_tags()
        eq_(tags, [])

    def test_get_source_url(self):
        table = Mock()
        item = MagicMock()
        item.__getitem__.return_value = 'booya.com/woot.gif'
        table.get_item.return_value = item
        image = Image('BADD00D')
        image._stored_table = table

        source_url = image.get_source_url()
        eq_(source_url, 'booya.com/woot.gif')

    def test_get_source_url__returns_none_if_no_such_image(self):
        table = Mock()
        table.get_item.side_effect = DynamoDBKeyNotFoundError('no such image')
        image = Image("TH15I5N0TEVENHEX")
        image._stored_table = table

        source_url = image.get_source_url()
        eq_(source_url, None)




#!/usr/bin/python

from setuptools import setup

setup(name="catsnap",
      version="2.0.0",
      description="catalog and store funny pictures",
      author="Andrew Lorente",
      author_email="andrew.lorente@gmail.com",
      url="github.com/andrewlorente/catsnap",
      packages=['catsnap',
                'catsnap.document',
                'catsnap.batch'],
      scripts=['scripts/catsnap'],
      install_requires=[
          "boto==2.5.2",
          "requests==0.13.2",
          "argparse==1.2.1",

          "mock==0.8",
          "nose==1.1.2",
          ],
      )
